# Copyright (c) 2023 Johnathan P. Irvin and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import json
import os
from argparse import ArgumentParser


def get_api_key() -> str:
    """
    Gets the API key from the environment variable OPENAI_API_KEY, or from the
    config.json file. If neither are found, a ValueError is raised. The config.json
    file should be in the following format:

    {
        "openai_api_key": "YOUR_API_KEY"
    }

    Raises:
        ValueError: If no API key is found.

    Returns:
        str: The API key.
    """
    api_key = os.environ.get("OPENAI_API_KEY")
    if api_key:
        return api_key

    candidates = ["config.json"]

    if 'SHELLCRAFT_CONFIG' in os.environ:
        candidates.append(os.environ['SHELLCRAFT_CONFIG'])

    config_dir = None
    if 'XDG_CONFIG_HOME' in os.environ:
        config_dir = os.path.join(os.environ['XDG_CONFIG_HOME'])
    elif 'HOME' in os.environ:
        config_dir = os.path.join(os.environ['HOME'], '.config')

    if config_dir:
        candidates.append(os.path.join(config_dir, 'shell-craft', 'config.json'))

    filename = None
    for candidate in candidates:
        if os.path.exists(candidate):
            filename = candidate
            break
    else:
        raise ValueError("No API key found")

    with open(filename) as f:
        config = json.load(f)

    api_key = config.get("openai_api_key")
    if api_key:
        return api_key

    raise ValueError("No API key found")

def add_arguments(parser: ArgumentParser):
    """
    Adds '--api-key' to the parser as an argument. If the argument is not
    provided, the API key is retrieved from the environment variable OPENAI_API_KEY,
    or from the config.json file. If neither are found, a ValueError is raised. The
    config.json file is the preferred method of providing the API key as it is a
    secret and should not be shared.

    Args:
        parser (ArgumentParser): The parser to add the argument to.
    """
    parser.add_argument(
        "--api-key",
        type=str,
        help="The OpenAI API key to use.",
    )

    try:
        if not parser.parse_known_args()[0].api_key:
            parser.set_defaults(api_key=get_api_key())
    except ValueError:
        print("No API key found. Please provide one with the --api-key argument.")
        parser.print_help()
        exit(1)



# Copyright (c) 2023 Johnathan P. Irvin and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from shell_craft import Service
from shell_craft.factories import PromptFactory

from .parser import PARSER, get_arguments


def main():
    """
    Main function that processes the command-line arguments and queries the
    OpenAI API using shell_craft.
    """
    args = get_arguments(PARSER)
    prompt = PromptFactory.get_prompt(args.prompt)

    if getattr(args, "refactor", False):
        prompt = prompt.refactoring
    elif getattr(args, "document", False):
        prompt = prompt.documentation
    elif getattr(args, "test", False):
        prompt = prompt.testing

    print(
        Service(
            api_key=args.api_key,
            prompt=prompt,
            model=args.model
        ).query(
            message=' '.join(args.request)
        )
    )



# Copyright (c) 2023 Johnathan P. Irvin and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from argparse import ArgumentParser
from shell_craft.factories import PromptFactory
from shell_craft.prompts import languages

ARGUMENTS = {
    'prompt': lambda x: PromptFactory.get_prompt(x) in languages.get_prompts()
}

def add_arguments(parser: ArgumentParser):
    """
    Adds 'request' argument to the parser. This argument is used to
    specify the input to prompt the API with. This argument is required,
    however, it can be piped in from another command or from a file.

    Args:
        parser (ArgumentParser): The parser to add the argument to.
    """    
    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument(
        '--refactor',
        action='store_true',
        help='Refactor the code.'
    )
    group.add_argument(
        '--document',
        action='store_true',
        help='Document the code.'
    )
    group.add_argument(
        '--test',
        action='store_true',
        help='Test the code.'
    )



# Copyright (c) 2023 Johnathan P. Irvin and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from dataclasses import dataclass
from typing import Optional

from .prompt import Prompt


@dataclass(frozen=True)
class LanguagePrompt(Prompt):
    refactoring: Optional[Prompt] = None
    documentation: Optional[Prompt] = None
    testing: Optional[Prompt] = None


def _generate_prompt(language: str) -> LanguagePrompt:
    return LanguagePrompt(
        content=" ".join(f"""
            You are {language}.
            You reply with valid {language}, nothing else.
            No explanations.
            You receive descriptions and return {language}.
        """.split()),
        refactoring=Prompt(
            content=" ".join(f"""
                You receive {language} and return {language}.
                You shorten the code.
                You remove bugs.
                You make the code more efficient.
            """.split())
        ),
        documentation=Prompt(
            content=" ".join(f"""
                You receive {language} and return {language} with document strings.
                You add documentation to each function.
                You add documentation to each class.
                You explain parameters and return values.
                You run the code with the documentation in place.
            """.split())
        ),
        testing=Prompt(
            content=" ".join(f"""
                You receive {language} and return {language} with tests.
                You write tests for each function.
                Tests are written in AAA (Arrange, Act, Assert) format.
                You do not repeat yourself.
                You do not return the original code.
            """.split())
        )
    )

def get_prompts() -> list[LanguagePrompt]:
    """
    Returns all of the language prompts. This is used to generate the
    language prompt choices. This is also used to define the exports.

    Returns:
        list[LanguagePrompt]: All of the language prompts.
    """    
    return [
        prompt for prompt
        in globals().values()
        if isinstance(prompt, LanguagePrompt)
    ]

BASH_PROMPT = _generate_prompt("Bash")
C_PROMPT = _generate_prompt("C")
C_SHARP_PROMPT = _generate_prompt("C#")
GO_PROMPT = _generate_prompt("Go")
POWERSHELL_PROMPT = _generate_prompt("PowerShell")
PYTHON_PROMPT = _generate_prompt("Python")
JAVA_PROMPT = _generate_prompt("Java")
JAVASCRIPT_PROMPT = _generate_prompt("JavaScript")


__all__ = get_prompts()



# Copyright (c) 2023 Johnathan P. Irvin and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import openai

from shell_craft.prompts import Prompt


class Service:
    def __init__(self, api_key: str, prompt: Prompt, model: str = "gpt-3.5-turbo"):
        """
        Initialize a new Service instance.

        Args:
            api_key (str): The OpenAI API key to use.
            prompt (Prompt): Informs the model how to act when generating a response.
            model (str, optional):
                The model to use when generating a response. Defaults to "gpt-3.5-turbo".
                Check out the OpenAI docs for more information on models and their
                pricing.
        """
        self._api_key: str = api_key
        self._prompt: Prompt = prompt
        self._model: str = model

    def query(self, message: str) -> str:
        """
        Query the model with a message.

        Args:
            message (str): The message to query the model with.

        Returns:
            str: The response from the model.
        """
        return openai.ChatCompletion.create(
            api_key=self._api_key,
            model=self._model,
            messages=self._prompt.messages + [
                {
                    "role": "user",
                    "content": message
                }
            ]
        )['choices'][0]['message']['content']
